unit uScene;

interface
uses
  Classes,
  System.UITypes,
  FMX.Graphics,
  System.Types;

type
  TSceneActionType = ( saBackColor, saFreehand, saFillColor, saStrokeColor, saClear, saUndo, saPenWidth );

  /// Represents a single drawing action.
  TSceneAction = record
    Action: TSceneActionType;
    Data: uint64;
    Points: array of TPointF;
    PointCount: uint32;
  end;

  TOnSceneActionEvent = procedure ( NewAction: TSceneAction ) of object;

  /// The scene stores all of the scene actions for transmission and
  /// rendering.
  TScene = class
  private
    fClearColor: TAlphaColor;
    fActions: array of TSceneAction;
    fActionCount: uint32;
    fActionInProgress: TSceneAction;
    fConstructing: boolean;
    fOnAction: TOnSceneActionEvent;
    fWidthAdjustment: Single;
    fHeightAdjustment: Single;
  private
    procedure RenderFreehand( Canvas: TCanvas; Action: TSceneAction );
    procedure RenderAction( Canvas: TCanvas; Action: TSceneAction );
  public
    procedure Clear;
    procedure BeginNewAction( ActionType: TSceneActionType; X, Y: Single );
    procedure UpdateAction( X, Y: single );
    procedure EndAction;
    procedure CancelAction;
    procedure AddAction(NewAction: TSceneAction);
    procedure Render( Canvas: TCanvas; ClipRect: TRectF );
  public
    class procedure ActionToStream( Action: TSceneAction; Stream: TStream );
    class procedure ActionFromStream( var Action: TSceneAction; Stream: TStream );
  public
    constructor Create; reintroduce;
    destructor Destroy; override;
  public
    ///
    ///  Event fires each time a new action is added to the scene.
    ///
    property OnAction: TOnSceneActionEvent read fOnAction write fOnAction;
  end;

implementation

const
  cActionGranularity = 512;
  cPointsGranularity = 512;

class procedure TScene.ActionFromStream(var Action: TSceneAction; Stream: TStream);
var
  i: int32;
  idx: int32;
begin
  //- Read action type.
  Stream.Read(I,sizeof(I));
  Action.Action := TSceneActionType(i);
  //- Read Data
  Stream.Read(I,sizeof(I));
  Action.Data := I;
  //- Read Points count.
  Stream.Read(I,sizeof(I));
  Action.PointCount := I;
  //- Load points.
  SetLength(Action.Points,I);
  Action.PointCount := I;
  if Action.PointCount=0 then begin
    exit;
  end;
  for idx := 0 to pred(Action.PointCount) do begin
    Stream.Read(Action.Points[idx].X,Sizeof(single));
    Stream.Read(Action.Points[idx].Y,Sizeof(single));
  end;
end;

class procedure TScene.ActionToStream(Action: TSceneAction; Stream: TStream);
var
  i: int32;
  idx: int32;
begin
  //- Write action
  i := int32(Action.Action);
  Stream.Write(I,sizeof(i));
  //- Write data.
  I := Action.Data;
  Stream.Write(I,Sizeof(I));
  //- Write point count
  i := Action.PointCount;
  Stream.Write(I,sizeof(i));
  //- Write points
  for idx := 0 to pred(Action.PointCount) do begin
    Stream.Write(Action.Points[idx].X,Sizeof(Single));
    Stream.Write(Action.Points[idx].Y,Sizeof(Single));
  end;
end;

procedure TScene.BeginNewAction( ActionType: TSceneActionType; X, Y: Single );
begin
  if fConstructing then begin
    exit;
  end;
  fConstructing := True;
  fActionInProgress.Action := ActionType;
  SetLength(fActionInProgress.Points,cPointsGranularity);
  fActionInProgress.PointCount := 0;
  Self.UpdateAction(X,Y);
end;

procedure TScene.CancelAction;
begin
  SetLength(fActionInProgress.Points,cPointsGranularity);
  fActionInProgress.PointCount := 0;
  fConstructing := False;
end;

procedure TScene.Clear;
begin
  SetLength(fActions,cActionGranularity);
  fActionCount := 0;
  fConstructing := False;
  fClearColor := TAlphaColorRec.Green;
end;

constructor TScene.Create;
begin
  inherited Create;
  fOnAction := nil;
  Clear;
end;

destructor TScene.Destroy;
begin
  SetLength(fActions,0);
  inherited Destroy;
end;

procedure TScene.AddAction( NewAction: TSceneAction );
var
  idx: uint32;
begin
  //- Back color is a special case...
  if NewAction.Action=saBackColor then begin
    fClearColor := TAlphaColor(NewAction.Data);
    if assigned(OnAction) then begin
      OnAction(NewAction);
    end;
    exit;
  end;
  //- Clear special case
  if NewAction.Action=saClear then begin
    Self.Clear;
    if assigned(OnAction) then begin
      OnAction(NewAction);
    end;
    exit;
  end;
  //- Undo special case
  if NewAction.Action=saUndo then begin
    if fActionCount>0 then begin
      fActionCount := fActionCount-1;
    end;
    if assigned(OnAction) then begin
      OnAction(NewAction);
    end;
    exit;
  end;
  //- Add the new action
  if fActionCount>=Length(fActions) then begin
    SetLength(fActions,Length(fActions)+cActionGranularity);
  end;
  fActions[fActionCount].Action := NewAction.Action;
  fActions[fActionCount].Data := NewAction.Data;
  fActions[fActionCount].PointCount := NewAction.PointCount;
  SetLength(fActions[fActionCount].Points,NewAction.PointCount);
  if NewAction.PointCount>0 then begin
    for idx := 0 to pred(NewAction.PointCount) do begin
      fActions[fActionCount].Points[idx].X := NewAction.Points[idx].X;
      fActions[fActionCount].Points[idx].Y := NewAction.Points[idx].Y;
    end;
  end;
  fActionCount := fActionCount + 1;
  //- if assigned, call the onaction event
  if assigned(OnAction) then begin
    OnAction(fActions[pred(fActionCount)]);
  end;
end;

procedure TScene.EndAction;
begin
  if not fConstructing then begin
    exit;
  end;
  fConstructing := False;
  AddAction(fActionInProgress);
end;

procedure TScene.Render(Canvas: TCanvas; ClipRect: TRectF);
var
  idx: uint32;
begin
  fWidthAdjustment := 1 / Canvas.Width;
  fHeightAdjustment := 1 / Canvas.Height;
  Canvas.Stroke.Kind := TBrushKind.Solid;
  Canvas.Fill.Kind := TBrushKind.Solid;
  Canvas.Stroke.Color := TAlphaColorRec.White;
  Canvas.Fill.Color := TAlphaColorRec.Blue;
  Canvas.ClearRect(ClipRect,fClearColor);
  //- Render existing actions.
  if fActionCount>0 then begin
    for idx := 0 to pred(fActionCount) do begin
      Self.RenderAction(Canvas,fActions[idx]);
    end;
  end;
  //- Render action in progress.
  if fConstructing then begin
    Self.RenderAction(Canvas,fActionInProgress);
  end;
end;

procedure TScene.RenderAction(Canvas: TCanvas; Action: TSceneAction);
begin
  case Action.Action of
    saPenWidth: Canvas.Stroke.Thickness := Action.Data;
    saFillColor: Canvas.Fill.Color := TAlphaColor(Action.Data);
    saStrokeColor: Canvas.Stroke.Color := TAlphaColor(Action.Data);
    saFreehand: RenderFreehand( Canvas, Action );
  end;
end;

procedure TScene.RenderFreehand(Canvas: TCanvas; Action: TSceneAction);
var
  idx: int32;
  StartPoint: TPointF;
  EndPoint: TPointF;
begin
  if Action.PointCount>1 then begin
    for idx := 1 to pred(Action.PointCount) do begin
      StartPoint := Action.Points[pred(idx)];
      EndPoint := Action.Points[idx];
      StartPoint.X := StartPoint.X / fWidthAdjustment;
      StartPoint.Y := StartPoint.Y / fHeightAdjustment;
      EndPoint.X := EndPoint.X / fWidthAdjustment;
      EndPoint.Y := EndPoint.Y / fHeightAdjustment;
      Canvas.DrawLine(StartPoint,EndPoint,1);
    end;
  end;
end;

procedure TScene.UpdateAction(X, Y: single);
begin
  if not fConstructing then begin
    exit;
  end;
  if fActionInProgress.PointCount>=Length(fActionInProgress.Points) then begin
    SetLength(fActionInProgress.Points,Length(fActionInProgress.Points)+cPointsGranularity);
  end;
  fActionInProgress.Points[fActionInProgress.PointCount].X := X * fWidthAdjustment;
  fActionInProgress.Points[fActionInProgress.PointCount].Y := Y * fHeightAdjustment;
  fActionInProgress.PointCount := fActionInProgress.PointCount + 1;
end;

end.
